#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List
import numpy as np
from gnes.helper import batch_iterator
from ..base import BaseImageEncoder
from ...helper import batching
from PIL import Image


class TFInceptionEncoder(BaseImageEncoder):

    def __init__(self, model_dir: str,
                 batch_size: int = 64,
                 select_layer: str = 'PreLogitsFlatten',
                 use_cuda: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.batch_size = batch_size
        self.select_layer = select_layer
        self._use_cuda = use_cuda
        self.inception_size_x = 299
        self.inception_size_y = 299

    def post_init(self):
        import tensorflow as tf
        from gnes.encoder.image.inception_cores.inception_v4 import inception_v4
        from gnes.encoder.image.inception_cores.inception_utils import inception_arg_scope

        arg_scope = inception_arg_scope()
        inception_v4.default_image_size = self.inception_size_x
        self.inputs = tf.placeholder(tf.float32, (None,
                                                  self.inception_size_x,
                                                  self.inception_size_y, 3))

        with tf.contrib.slim.arg_scope(arg_scope):
            self.logits, self.end_points = inception_v4(self.inputs,
                                                        is_training=False,
                                                        dropout_keep_prob=1.0)

        config = tf.ConfigProto(log_device_placement=False)
        if self._use_cuda:
            config.gpu_options.allow_growth = True
        self.sess = tf.Session(config=config)
        self.saver = tf.train.Saver()
        self.saver.restore(self.sess, self.model_dir)

    @batching
    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        ret = []
        img = [(np.array(Image.fromarray(im).resize((self.inception_size_x,
                         self.inception_size_y)), dtype=np.float32) * 2 / 255. - 1.) for im in img]
        for _im in batch_iterator(img, self.batch_size):
            _, end_points_ = self.sess.run((self.logits, self.end_points),
                                           feed_dict={self.inputs: _im})
            ret.append(end_points_[self.select_layer])
        return np.concatenate(ret, axis=0).astype(np.float32)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=low-comment-ratio


from typing import List, Any, Union, Dict, Callable

import numpy as np

from ..base import TrainableBase
from ..proto import gnes_pb2


class BaseEncoder(TrainableBase):

    def encode(self, data: Any, *args, **kwargs) -> Any:
        pass

    def _copy_from(self, x: 'BaseEncoder') -> None:
        pass


class BaseImageEncoder(BaseEncoder):

    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pass


class BaseTextEncoder(BaseEncoder):

    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        pass


class BaseNumericEncoder(BaseEncoder):

    def encode(self, text: np.ndarray, *args, **kwargs) -> np.ndarray:
        pass


class BaseBinaryEncoder(BaseEncoder):

    def encode(self, data: np.ndarray, *args, **kwargs) -> bytes:
        if data.dtype != np.uint8:
            raise ValueError('data must be np.uint8 but received %s' % data.dtype)
        return data.tobytes()


class CompositionalEncoder(BaseEncoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._component = None  # type: List['BaseEncoder']

    @property
    def component(self) -> Union[List['BaseEncoder'], Dict[str, 'BaseEncoder']]:
        return self._component

    @property
    def is_pipeline(self):
        return isinstance(self.component, list)

    @component.setter
    def component(self, comps: Callable[[], Union[list, dict]]):
        if not callable(comps):
            raise TypeError('component mus be a callable function that returns '
                            'a List[BaseEncoder]')
        if not getattr(self, 'init_from_yaml', False):
            self._component = comps()
        else:
            self.logger.info('component is omitted from construction, '
                             'as it is initialized from yaml config')

    def close(self):
        super().close()
        # pipeline
        if isinstance(self.component, list):
            for be in self.component:
                be.close()
        # no typology
        elif isinstance(self.component, dict):
            for be in self.component.values():
                be.close()
        elif self.component is None:
            pass
        else:
            raise TypeError('component must be dict or list, received %s' % type(self.component))

    def _copy_from(self, x: 'CompositionalEncoder'):
        if isinstance(self.component, list):
            for be1, be2 in zip(self.component, x.component):
                be1._copy_from(be2)
        elif isinstance(self.component, dict):
            for k, v in self.component.items():
                v._copy_from(x.component[k])
        else:
            raise TypeError('component must be dict or list, received %s' % type(self.component))

    @classmethod
    def to_yaml(cls, representer, data):
        tmp = super()._dump_instance_to_yaml(data)
        tmp['component'] = data.component
        return representer.represent_mapping('!' + cls.__name__, tmp)

    @classmethod
    def from_yaml(cls, constructor, node):
        obj, data = super()._get_instance_from_yaml(constructor, node)
        if 'component' in data:
            obj.component = lambda: data['component']
        return obj


class PipelineEncoder(CompositionalEncoder):
    def encode(self, data: Any, *args, **kwargs) -> Any:
        if not self.component:
            raise NotImplementedError
        for be in self.component:
            data = be.encode(data, *args, **kwargs)
        return data

    def train(self, data, *args, **kwargs):
        if not self.component:
            raise NotImplementedError
        for idx, be in enumerate(self.component):
            be.train(data, *args, **kwargs)
            if idx + 1 < len(self.component):
                data = be.encode(data, *args, **kwargs)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..base import BasePreprocessor
from ...proto import gnes_pb2
from typing import List


class BaseImagePreprocessor(BasePreprocessor):
    doc_type = gnes_pb2.Document.IMAGE

    def __init__(self,
                 target_img_size: int = 224,
                 is_rgb: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_img_size = target_img_size
        self.is_rgb = is_rgb

    def _get_all_chunks_weight(self, image_set: List['np.ndarray']) -> List[float]:
        pass

    @classmethod
    def _torch_transform(cls, image):
        import torchvision.transforms as transforms
        return transforms.Compose([transforms.ToTensor(),
                                transforms.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5))])(image)


#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=low-comment-ratio

import threading
import uuid
from concurrent import futures

import grpc
import zmq

from .base import build_socket
from ..helper import set_logger
from ..proto import gnes_pb2, gnes_pb2_grpc, send_message, recv_message

__all__ = ['GRPCFrontend']


class ZmqContext(object):
    """The zmq context class."""

    def __init__(self, args):
        """Database connection context.

        Args:
            servers: a list of config dicts for connecting to database
            dbapi_name: the name of database engine
        """
        self.args = args

        self.tlocal = threading.local()
        self.tlocal.client = None

    def __enter__(self):
        """Enter the context."""
        client = ZmqClient(self.args)
        self.tlocal.client = client
        return client

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit the context."""
        self.tlocal.client.close()
        self.tlocal.client = None


class ZmqClient:

    def __init__(self, args):
        self.args = args
        self.identity = args.identity if 'identity' in args else None
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)
        self.ctx = zmq.Context()
        self.ctx.setsockopt(zmq.LINGER, 0)
        self.receiver, _ = build_socket(self.ctx, self.args.host_in, self.args.port_in, self.args.socket_in,
                                        getattr(self, 'identity', None))
        self.sender, _ = build_socket(self.ctx, self.args.host_out, self.args.port_out, self.args.socket_out,
                                      getattr(self, 'identity', None))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.sender.close()
        self.receiver.close()
        self.ctx.term()

    def send_message(self, message: "gnes_pb2.Message", timeout: int = -1):
        send_message(self.sender, message, timeout=timeout)

    def recv_message(self, timeout: int = -1) -> gnes_pb2.Message:
        return recv_message(self.receiver, timeout=timeout)


class GNESServicer(gnes_pb2_grpc.GnesRPCServicer):

    def __init__(self, args):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, args.verbose)
        self.zmq_context = ZmqContext(args)

    def add_envelope(self, body: 'gnes_pb2.Request', zmq_client: 'ZmqClient'):
        msg = gnes_pb2.Message()
        msg.envelope.client_id = zmq_client.identity if zmq_client.identity else ''
        if body.request_id:
            msg.envelope.request_id = body.request_id
        else:
            msg.envelope.request_id = str(uuid.uuid4())
            self.logger.warning('request_id is missing, filled it with a random uuid!')
        msg.envelope.part_id = 1
        msg.envelope.num_part = 1
        msg.envelope.timeout = 5000
        r = msg.envelope.routes.add()
        r.service = zmq_client.__class__.__name__
        r.timestamp.GetCurrentTime()
        msg.request.CopyFrom(body)
        return msg

    def _Call(self, request, context):
        self.logger.info('received a new request: %s' % request.request_id or 'EMPTY_REQUEST_ID')
        with self.zmq_context as zmq_client:
            msg = self.add_envelope(request, zmq_client)
            zmq_client.send_message(msg, self.args.timeout)
            resp = zmq_client.recv_message(self.args.timeout)
            self.logger.info("received message done!")
            return resp.response

    def Train(self, request, context):
        return self._Call(request, context)

    def Index(self, request, context):
        return self._Call(request, context)

    def Search(self, request, context):
        return self._Call(request, context)


class GRPCFrontend:
    def __init__(self, args):
        self.logger = set_logger(self.__class__.__name__, args.verbose)
        self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=args.max_concurrency),
            options=[('grpc.max_send_message_length', args.max_send_size*1024*1024),
                     ('grpc.max_receive_message_length', args.max_receive_size*1024*1024)])
        self.logger.info('start a grpc server with %d workers' % args.max_concurrency)
        gnes_pb2_grpc.add_GnesRPCServicer_to_server(GNESServicer(args), self.server)

        # Start GRPC Server
        self.bind_address = '{0}:{1}'.format(args.grpc_host, args.grpc_port)
        self.server.add_insecure_port(self.bind_address)

    def __enter__(self):
        self.server.start()
        self.logger.info('grpc service is listening at: %s' % self.bind_address)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



from typing import Generator

from ..base import BaseMapRouter
from ...helper import batch_iterator
from ...proto import gnes_pb2


class PublishRouter(BaseMapRouter):

    def __init__(self, num_part: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_part = num_part

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        msg.envelope.num_part = self.num_part
        yield msg


class DocBatchRouter(BaseMapRouter):
    def __init__(self, batch_size: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.batch_size = batch_size

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        if self.batch_size and self.batch_size > 0:
            batches = [b for b in batch_iterator(msg.request.index.docs, self.batch_size)]
            num_part = len(batches)
            for p_idx, b in enumerate(batches, start=1):
                _msg = gnes_pb2.Message()
                _msg.CopyFrom(msg)
                _msg.request.index.ClearField('docs')
                _msg.request.index.docs.extend(b)
                _msg.envelope.part_id = p_idx
                _msg.envelope.num_part = num_part
                yield _msg



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=low-comment-ratio


import inspect
import os
import pickle
import tempfile
import uuid
from functools import wraps
from typing import Dict, Any, Union, TextIO, TypeVar, Type

import ruamel.yaml.constructor

from ..helper import set_logger, profiling, yaml, parse_arg, touch_dir, FileLock

__all__ = ['TrainableBase']

T = TypeVar('T', bound='TrainableBase')


def register_all_class(cls2file_map: Dict):
    for k in cls2file_map:
        try:
            import_class_by_str(k)
        except ImportError:
            pass


def import_class_by_str(name: str):
    def _import(module_name, class_name):
        import importlib

        cls2file = getattr(importlib.import_module('gnes.%s' % module_name), '_cls2file_map')
        if class_name in cls2file:
            return getattr(importlib.import_module('gnes.%s.%s' % (module_name, cls2file[class_name])), class_name)

    search_modules = ['encoder', 'indexer', 'preprocessor', 'router']

    for m in search_modules:
        r = _import(m, name)
        if r:
            return r
    else:
        raise ImportError('Can not locate any class with name: %s, misspelling?' % name)


class TrainableType(type):
    default_property = {
        'is_trained': False,
        'batch_size': None,
    }

    def __new__(cls, *args, **kwargs):
        _cls = super().__new__(cls, *args, **kwargs)
        return cls.register_class(_cls)

    def __call__(cls, *args, **kwargs):
        # do _preload_package
        getattr(cls, '_pre_init', lambda *x: None)()

        obj = type.__call__(cls, *args, **kwargs)

        # set attribute
        for k, v in TrainableType.default_property.items():
            if not hasattr(obj, k):
                setattr(obj, k, v)

        # do _post_init()
        getattr(obj, '_post_init_wrapper', lambda *x: None)()
        return obj

    @staticmethod
    def register_class(cls):
        # print('try to register class: %s' % cls.__name__)
        reg_cls_set = getattr(cls, '_registered_class', set())
        if cls.__name__ not in reg_cls_set:
            # print('reg class: %s' % cls.__name__)
            cls.__init__ = TrainableType._store_init_kwargs(cls.__init__)
            if os.environ.get('GNES_PROFILING', False):
                for f_name in ['train', 'encode', 'add', 'query']:
                    if getattr(cls, f_name, None):
                        setattr(cls, f_name, profiling(getattr(cls, f_name)))

            if getattr(cls, 'train', None):
                setattr(cls, 'train', TrainableType._as_train_func(getattr(cls, 'train')))

            reg_cls_set.add(cls.__name__)
            setattr(cls, '_registered_class', reg_cls_set)
            yaml.register_class(cls)
        return cls

    @staticmethod
    def _as_train_func(func):
        @wraps(func)
        def arg_wrapper(self, *args, **kwargs):
            if self.is_trained:
                self.logger.warning('"%s" has been trained already, '
                                    'training it again will override the previous training' % self.__class__.__name__)
            f = func(self, *args, **kwargs)
            self.is_trained = True
            return f

        return arg_wrapper

    @staticmethod
    def _store_init_kwargs(func):
        @wraps(func)
        def arg_wrapper(self, *args, **kwargs):
            taboo = {'self', 'args', 'kwargs'}
            all_pars = inspect.signature(func).parameters
            tmp = {k: v.default for k, v in all_pars.items() if k not in taboo}
            tmp_list = [k for k in all_pars.keys() if k not in taboo]
            # set args by aligning tmp_list with arg values
            for k, v in zip(tmp_list, args):
                tmp[k] = v
            # set kwargs
            for k, v in kwargs.items():
                if k in tmp:
                    tmp[k] = v

            if self.store_args_kwargs:
                if args: tmp['args'] = args
                if kwargs: tmp['kwargs'] = kwargs

            if getattr(self, '_init_kwargs_dict', None):
                self._init_kwargs_dict.update(tmp)
            else:
                self._init_kwargs_dict = tmp
            f = func(self, *args, **kwargs)
            return f

        return arg_wrapper


class TrainableBase(metaclass=TrainableType):
    store_args_kwargs = False
    lock_work_dir = False

    def __init__(self, *args, **kwargs):
        self.is_trained = False
        self._obj_id = str(uuid.uuid4()).split('-')[0]
        self._obj_pickle_name = '%s%s.bin' % (self.__class__.__name__, self._obj_id)
        self._obj_yaml_name = '%s%s.yml' % (self.__class__.__name__, self._obj_id)
        self._work_dir = os.getcwd()
        self.verbose = 'verbose' in kwargs and kwargs['verbose']
        self.logger = set_logger(self.__class__.__name__, self.verbose)
        self._post_init_vars = set()

    def _post_init_wrapper(self):
        _before = set(list(self.__dict__.keys()))
        self.post_init()
        self._post_init_vars = {k for k in self.__dict__ if k not in _before}

    def post_init(self):
        pass

    @classmethod
    def pre_init(cls):
        pass

    @property
    def pickle_full_path(self):
        return os.path.join(self.work_dir, self._obj_pickle_name)

    @property
    def yaml_full_path(self):
        return os.path.join(self.work_dir, self._obj_yaml_name)

    @property
    def work_dir(self):
        return self._work_dir

    @work_dir.setter
    def work_dir(self, value: str):
        touch_dir(value)
        if self.lock_work_dir:
            self._file_lock = FileLock(os.path.join(value, "LOCK"))
            if self._file_lock.acquire() is None:
                raise RuntimeError(
                    "this model\'s work_dir %r is used and locked by another model" %
                    value)
        self._work_dir = value

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['logger']
        if '_file_lock' in d:
            del d['_file_lock']
        for k in self._post_init_vars:
            del d[k]
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.logger = set_logger(self.__class__.__name__, self.verbose)
        if self.lock_work_dir:
            # trigger the lock again
            self.work_dir = self._work_dir
        try:
            self._post_init_wrapper()
        except ImportError:
            self.logger.info('ImportError is often caused by a missing component, '
                             'which often can be solved by "pip install" relevant package.')

    def train(self, *args, **kwargs):
        pass

    @profiling
    def dump(self, filename: str = None) -> None:
        f = filename or self.pickle_full_path
        if not f:
            f = tempfile.NamedTemporaryFile('w', delete=False, dir=os.environ.get('NES_TEMP_DIR', None)).name
        with open(f, 'wb') as fp:
            pickle.dump(self, fp)
        self.logger.info('model is pickled to %s' % f)

    @profiling
    def dump_yaml(self, filename: str = None) -> None:
        f = filename or self.yaml_full_path
        if not f:
            f = tempfile.NamedTemporaryFile('w', delete=False, dir=os.environ.get('NES_TEMP_DIR', None)).name
        with open(filename, 'w') as fp:
            yaml.dump(self, fp)
        self.logger.info('model\'s yaml config is dump to %s' % f)

    @classmethod
    def load_yaml(cls: Type[T], filename: Union[str, TextIO]) -> T:
        if not filename: raise FileNotFoundError
        if isinstance(filename, str):
            with open(filename) as fp:
                return yaml.load(fp)
        else:
            with filename:
                return yaml.load(filename)

    @staticmethod
    @profiling
    def load(filename: str) -> T:
        if not filename: raise FileNotFoundError
        with open(filename, 'rb') as fp:
            return pickle.load(fp)

    def close(self):
        if self.lock_work_dir:
            self._file_lock.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def _get_tags_from_node(node):
        def node_recurse_generator(n):
            if n.tag.startswith('!'):
                yield n.tag.lstrip('!')
            for nn in n.value:
                if isinstance(nn, tuple):
                    for k in nn:
                        yield from node_recurse_generator(k)
                elif isinstance(nn, ruamel.yaml.nodes.Node):
                    yield from node_recurse_generator(nn)

        return list(set(list(node_recurse_generator(node))))

    @classmethod
    def to_yaml(cls, representer, data):
        tmp = data._dump_instance_to_yaml(data)
        return representer.represent_mapping('!' + cls.__name__, tmp)

    @classmethod
    def from_yaml(cls, constructor, node, stop_on_import_error=False):
        return cls._get_instance_from_yaml(constructor, node, stop_on_import_error)[0]

    @classmethod
    def _get_instance_from_yaml(cls, constructor, node, stop_on_import_error=False):
        try:
            for c in cls._get_tags_from_node(node):
                import_class_by_str(c)
        except ImportError as ex:
            if stop_on_import_error:
                raise RuntimeError('Cannot import module, pip install may required') from ex

        data = ruamel.yaml.constructor.SafeConstructor.construct_mapping(
            constructor, node, deep=True)
        cls.init_from_yaml = True

        if cls.store_args_kwargs:
            p = data.get('parameter', {})  # type: Dict[str, Any]
            a = p.pop('args') if 'args' in p else ()
            k = p.pop('kwargs') if 'kwargs' in p else {}
            # maybe there are some hanging kwargs in "parameter"
            tmp_a = (cls._convert_env_var(v) for v in a)
            tmp_p = {kk: cls._convert_env_var(vv) for kk, vv in {**k, **p}.items()}
            obj = cls(*tmp_a, **tmp_p)
        else:
            tmp_p = {kk: cls._convert_env_var(vv) for kk, vv in data.get('parameter', {}).items()}
            obj = cls(**tmp_p)

        for k, v in data.get('property', {}).items():
            old = getattr(obj, k, None)
            setattr(obj, k, v)
            if old and old != v:
                obj.logger.info('property: %r is replaced from %r to %r' % (k, old, v))

        cls.init_from_yaml = False

        return obj, data

    @staticmethod
    def _convert_env_var(v):
        if isinstance(v, str):
            return parse_arg(os.path.expandvars(v))
        else:
            return v

    @staticmethod
    def _dump_instance_to_yaml(data):
        # note: we only dump non-default property for the sake of clarity
        p = {k: getattr(data, k) for k, v in TrainableType.default_property.items() if getattr(data, k) != v}
        a = {k: v for k, v in data._init_kwargs_dict.items()}
        r = {}
        if a:
            r['parameter'] = a
        if p:
            r['property'] = p
        return r



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=low-comment-ratio

import io
import subprocess as sp
from typing import List, Callable

import cv2
import numpy as np
from PIL import Image
import imagehash


def get_video_frames(buffer_data: bytes, image_format: str = "cv2",
                     **kwargs) -> List["np.ndarray"]:
    ffmpeg_cmd = ['ffmpeg', '-i', '-', '-f', 'image2pipe']

    # example k,v pair:
    #    (-s, 420*360)
    #    (-vsync, vfr)
    #    (-vf, select=eq(pict_type\,I))
    for k, v in kwargs.items():
        ffmpeg_cmd.append('-' + k)
        ffmpeg_cmd.append(v)

    # (-c:v, png) output bytes in png format
    # (-an, -sn) disable audio processing
    # (-) output to stdout pipeline
    ffmpeg_cmd += ['-c:v', 'png', '-an', '-sn', '-']

    with sp.Popen(
            ffmpeg_cmd, stdin=sp.PIPE, stdout=sp.PIPE, bufsize=-1,
            shell=False) as pipe:
        stream, _ = pipe.communicate(buffer_data)

        # raw bytes for multiple PNGs.
        # split by PNG EOF b'\x89PNG'
        stream = stream.split(b'\x89PNG')

        if len(stream) <= 1:
            return []

        # reformulate the full pngs for feature processings.
        if image_format == 'pil':
            frames = [
                Image.open(io.BytesIO(b'\x89PNG' + _)) for _ in stream[1:]
            ]
        elif image_format == 'cv2':
            frames = [
                cv2.imdecode(np.frombuffer(b'\x89PNG' + _, np.uint8), 1)
                for _ in stream[1:]
            ]
        else:
            raise NotImplementedError

    return frames


def block_descriptor(image: "np.ndarray",
                     descriptor_fn: Callable,
                     num_blocks: int = 3) -> "np.ndarray":
    h, w, _ = image.shape    # find shape of image and channel
    block_h = int(np.ceil(h / num_blocks))
    block_w = int(np.ceil(w / num_blocks))

    descriptors = []
    for i in range(0, h, block_h):
        for j in range(0, w, block_w):
            block = image[i:i + block_h, j:j + block_w]
            descriptors.extend(descriptor_fn(block))

    return np.array(descriptors)


def pyramid_descriptor(image: "np.ndarray",
                       descriptor_fn: Callable,
                       max_level: int = 2) -> "np.ndarray":
    descriptors = []
    for level in range(max_level + 1):
        num_blocks = 2**level
        descriptors.extend(block_descriptor(image, descriptor_fn, num_blocks))

    return np.array(descriptors)


def rgb_histogram(image: "np.ndarray") -> "np.ndarray":
    _, _, c = image.shape
    hist = [
        cv2.calcHist([image], [i], None, [256], [0, 256]) for i in range(c)
    ]
    # normalize hist
    hist = np.array([h / np.sum(h) for h in hist]).flatten()
    return hist


def hsv_histogram(image: "np.ndarray") -> "np.ndarray":
    _, _, c = image.shape
    hsv = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)

    # sizes = [180, 256, 256]
    # ranges = [(0, 180), (0, 256), (0, 256)]

    # hist = [
    #     cv2.calcHist([hsv], [i], None, [sizes[i]], ranges[i]) for i in range(c)
    # ]

    hist = [cv2.calcHist([hsv], [i], None, [256], [0, 256]) for i in range(c)]
    # normalize hist
    hist = np.array([h / np.sum(h) for h in hist]).flatten()
    return hist


def phash_descriptor(image: "np.ndarray") -> "imagehash.ImageHash":
    image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
    return imagehash.phash(image)


def compute_descriptor(image: "np.ndarray",
                       method: str = "rgb_histogram",
                       **kwargs) -> "np.array":
    funcs = {
        'rgb_histogram': rgb_histogram,
        'hsv_histogram': hsv_histogram,
        'block_rgb_histogram': lambda image: block_descriptor(image, rgb_histogram, kwargs.get("num_blocks", 3)),
        'block_hsv_histogram': lambda image: block_descriptor(image, hsv_histogram, kwargs.get("num_blocks", 3)),
        'pyramid_rgb_histogram': lambda image: pyramid_descriptor(image, rgb_histogram, kwargs.get("max_level", 2)),
        'pyramid_hsv_histogram': lambda image: pyramid_descriptor(image, hsv_histogram, kwargs.get("max_level", 2)),
    }
    return funcs[method](image)


def compare_descriptor(descriptor1: "np.ndarray",
                       descriptor2: "np.ndarray",
                       metric: str = "chisqr") -> float:
    dist_metric = {
        "correlation": cv2.HISTCMP_CORREL,
        "chisqr": cv2.HISTCMP_CHISQR,
        "chisqr_alt": cv2.HISTCMP_CHISQR_ALT,
        "intersection": cv2.HISTCMP_INTERSECT,
        "bhattacharya": cv2.HISTCMP_BHATTACHARYYA,
        "hellinguer": cv2.HISTCMP_HELLINGER,
        "kl_div": cv2.HISTCMP_KL_DIV
    }

    return cv2.compareHist(descriptor1, descriptor2, dist_metric[metric])



from .base import BaseImagePreprocessor
from ...proto import array2blob
from PIL import Image
import numpy as np
import io
import os


class SegmentPreprocessor(BaseImagePreprocessor):

    def __init__(self, model_name: str,
                 model_dir: str,
                 target_img_size: int = 224,
                 _use_cuda: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name
        self.model_dir = model_dir
        self.target_img_size = target_img_size
        self.model_name = model_name
        self._use_cuda = _use_cuda

    def post_init(self):
        import torch
        import torchvision.models as models

        os.environ['TORCH_HOME'] = self.model_dir
        self._model = getattr(models.detection, self.model_name)(pretrained=True)
        self._model = self._model.eval()
        if self._use_cuda:
            # self._model.cuda()
            self._device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            self._model = self._model.to(self._device)

    def apply(self, doc: 'gnes_pb2.Document'):
        super().apply(doc)
        if doc.raw_bytes:
            original_image = Image.open(io.BytesIO(doc.raw_bytes))
            image_tensor = self._torch_transform(original_image)

            seg_output = self._model([image_tensor])
            chunks = seg_output[0]['boxes'].tolist()
            weight = seg_output[0]['scores'].tolist()
            for ci, ele in enumerate(zip(chunks, weight)):
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.blob.CopyFrom(array2blob(self._crop_image_reshape(original_image, ele[0])))
                c.offset_1d = ci
                c.weight = ele[1]
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    def _crop_image_reshape(self, original_image, coordinates):
        return np.array(original_image.crop(coordinates).resize((self.target_img_size,
                                                                 self.target_img_size)))



import zipfile

import grpc

from ..proto import gnes_pb2_grpc, RequestGenerator


class CLIClient:
    def __init__(self, args):
        if args.data_type == 'text':
            all_bytes = [v.encode() for v in args.txt_file]
        elif args.data_type == 'image':
            zipfile_ = zipfile.ZipFile(args.image_zip_file, "r")
            all_bytes = [zipfile_.open(v).read() for v in zipfile_.namelist()]
        elif args.data_type == 'video':
            raise NotImplementedError
        else:
            raise ValueError(
                '--data_type can only be text, image or video')

        with grpc.insecure_channel(
                '%s:%s' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                         ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)

            if args.mode == 'train':
                for req in RequestGenerator.train(all_bytes, args.batch_size):
                    resp = stub._Call(req)
                    print(resp)
            elif args.mode == 'index':
                for req in RequestGenerator.index(all_bytes, args.batch_size):
                    resp = stub._Call(req)
                    print(resp)
            elif args.mode == 'query':
                for idx, q in enumerate(all_bytes):
                    for req in RequestGenerator.query(q, args.top_k):
                        resp = stub._Call(req)
                        print(resp)
                        print('query %d result: %s' % (idx, resp))
                        input('press any key to continue...')
